/*
  pedometer - a tools to retreive data and manage settings from Crivit Z31192 pedometer

  Copyright (c) 2013, Laurent Mazet <mazet@sftndesign.org>
*/

/*
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/* depend: */
/* cflags: -Wno-missing-field-initializers */
/* linker: -lusb-1.0 -ldb */

#include <assert.h>
#include <db.h>
#include <getopt.h>
#include <libusb-1.0/libusb.h>
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>

/* version */

#define VERSION "1.2"

#include "version.h"

/* gobal variables */

char *progname = NULL;
char dbfilename[1024] = { 0 };

/* system type */

typedef enum {
    system_undef_e = 0,
    system_metric_e = 1,
    system_imperial_e = 2,
    system_unknown_e = 3
} system_t;

/* parameter structure type */

typedef struct {
    int gait; /* in meter or feet */
    system_t system; /* 1 for metric, 0 for imperial */
    int target; /* in step */
    int weight; /* in kg or pound */
} parameter_t;

/* data type */

typedef struct {
    int calories;
    char date[12];
    int distance;
    int steps;
    int walking_time;
} data_t;

/* configuration structure type */
typedef struct {
    system_t system;
} config_t;

/* verbose macros */

int verbose = 1;

#define ERROR 0
#define WARNING 1
#define DEBUG 2

#define VERBOSE(level, statement...)\
    do {                                                         \
        if (level <= verbose) {                                  \
            printf ("\033[1;%dm", (level == ERROR) ? 31 :        \
                                  (level == WARNING) ? 33 : 34); \
            statement;                                           \
            printf ("\033[0;0m");                                \
        }                                                        \
    } while(0)

/* Divide and round */

#define DIVROUND(x, y) ((x * y + 5000) / 10000)

/* dump function */

void dump (char* str, unsigned char *data, int size)
{

    printf ("%s:", str);

    int i = 0;
    while (i < size) {
        printf (" %02X", data[i++]);
        if (i % 8 == 0)
	        printf ("\n");
    }
    if (i % 8 != 0)
        printf ("\n");
}

/* transaction function */

void transact (libusb_device_handle * dev, unsigned char *data, int tx_len, unsigned char *capture, int rx_len)
{

    VERBOSE (DEBUG, dump ("Sending ", data, tx_len));

    int r = libusb_control_transfer (dev, LIBUSB_REQUEST_TYPE_CLASS | LIBUSB_RECIPIENT_INTERFACE,
                                     0x09, 0x200, 0x0, data, tx_len, 5000);

    if (r != 8) {
        VERBOSE (ERROR, printf ("Write Error %d\n", r));
        return;
    }

    int actual = 0;

    r = libusb_bulk_transfer (dev, 0x81, capture, rx_len, &actual, 2000);

    VERBOSE (DEBUG, dump ("Received", capture, 6));

    if (r != 0)
        VERBOSE (ERROR, printf ("Read error %d. Actual read %d\n", r, actual));
}

/* pack function */

void packit (unsigned char *data, char* str)
{
    /*
    ** Converts a supplied string 'str' into BCD in the array pointed to by data
    ** No checks are made on the supplied string to make sure each digit is numeric.
    */

    unsigned char *p1 = data;
    char *p2 = str;
    unsigned char x = 0;
    int char_count = 0;

    while (*p2 != 0) {
        x += *p2++ - '0';
        char_count++;
        if ((char_count & 1) != 0) {
            x <<= 4;
        } else {
            *p1++ = x;
            x = 0;
        }
    }
}

/* unpack function */

int unpackit (unsigned char *data, int nibble, int length)
{
    /*
    ** say data holds 00 07 20 85 01 f2 01 12
    ** nibble = 3, length = 2, returns 72
    ** nibble = 0, length = 3, returns 0
    */

    int result = 0;
    int part;
    unsigned char *p1;

    if (length < 0) return 0;

    p1 = data + (nibble / 2);

    while (length--) {
        if ((nibble & 1) != 0) {
            part = *p1 & 0x0F;
            p1++;
        } else {
            part = (*p1 & 0xF0) >> 4;
        }
        result = result * 10 + part;
        nibble++;
    }

    return result;
}

/* find device */

libusb_device_handle *find_device (void)
{
    /*
    ** Returns pointer to the device if the Pedometer is online, else
    ** NULL If you only call this routine once, i.e. during program
    ** initiation, then unplug the pedometer, and plug it in again,
    ** something gets screwed up and although the reads appear to
    ** work, the data is a bit iffy.
    ** TODO: Look into why this is.
    */

    libusb_device **devs; //pointer to pointer of device, used to retrieve a list of devices
    libusb_context *ctx = NULL; //a libusb session

    int r = libusb_init (&ctx); //initialize the library for the session we just declared
    if (r < 0) {
        VERBOSE (ERROR, printf ("Init Error %d\n", r)); //there was an error
        return NULL;
    }
    libusb_set_debug (ctx, 3); //set verbosity level to 3, as suggested in the documentation

    ssize_t cnt = libusb_get_device_list (ctx, &devs); //get the list of devices
    if (cnt < 0) {
        VERBOSE (ERROR, printf ("Get Device Error\n")); //there was an error
        return NULL;
    }
    VERBOSE (DEBUG, printf ("List of devices: %d\n", (int)cnt));

    libusb_device_handle *dev = libusb_open_device_with_vid_pid (ctx, 0x0C45, 0x70C1); //these are vendorID and productID I found for my usb device
    libusb_free_device_list (devs, 1); //free the list, unref the devices in it
    if (dev == NULL) {
        VERBOSE (ERROR, printf ("Cannot open device\n"));
        return NULL;
    }

    if (libusb_kernel_driver_active (dev, 0) == 1) { //find out if kernel driver is attached
        VERBOSE (WARNING, printf ("Kernel Driver Active\n"));

        if (libusb_detach_kernel_driver(dev, 0) == 0) { //detach it
            VERBOSE (WARNING, printf ("Kernel Driver Detached!\n"));
        }
    }

    r = libusb_claim_interface (dev, 0); //claim interface 0

    if (r < 0) {
        VERBOSE (ERROR, printf ("Cannot Claim Interface\n"));
        return NULL;
    }

    return dev;
}

/* get parameters */

int get_parameters (libusb_device_handle *dev, parameter_t *p)
{

    if (dev == NULL)
        return 0;

    unsigned char data[8];
    unsigned char capture[12];

    memcpy (data, (const char[8]){ 0, 0, 0, 0, 0, 0, 0, 2 }, 8);
    transact (dev, data, 8, capture, 8);

    memcpy (data, (const char[8]){ 6, 0, 0, 0, 0, 0, 0, 2 }, 8);
    transact (dev, data, 8, capture + 6, 8);

    /*
    ** Decide if metric or not. I've seen this byte 0x00, 0x01, 0x04 and 0x05.
    ** Cases 1 & 5 were metric
    */

    p->system = (capture[4] & 1) ? system_metric_e : system_imperial_e;
    VERBOSE (DEBUG, printf ("Pedometer metric or imperial: %d\n", p->system));

    p->weight = unpackit (capture, 2, 3);
    VERBOSE (DEBUG, printf ("Pedometer weight: %d\n", p->weight));

    p->gait = unpackit (capture, 5, 3);
    VERBOSE (DEBUG, printf ("Pedometer gait: %d\n", p->gait));

    /*
    ** The current step target is spread over two USB transactions
    */

    p->target = unpackit (capture, 21, 3) * 100 + unpackit (capture, 0, 2);
    VERBOSE (DEBUG, printf ("Pedometer target: %d\n", p->target));

    return 1;
}

/* set parameters */

int set_parameters (libusb_device_handle *dev, parameter_t *p)
{

    if (dev == NULL)
        return 0;

    unsigned char data[8];
    unsigned char capture[6];

    /* set gait, system and weight (%10) */

    memcpy (data, (const char[8]){ 1, 0, 0, 0, 0, 0, 0, 2 }, 8);
    transact (dev, data, 8, capture, 8);

    char buffer[17] = { 0 };
    snprintf (buffer, sizeof (buffer), "0000%03d%03d%02d0103", p->weight % 10, p->gait,
              4 /* allow clock setting */ + ((p->system == system_metric_e) ? 1 : 0));
    packit (data, buffer);
    transact (dev, data, 8, capture, 8);

    /* set time, target and weight (/10) */

    memcpy (data, (const char[8]){ 4, 0, 0, 0, 0, 0, 0, 2 }, 8);
    transact (dev, data, 8, capture, 8);

    time_t now = time (0);
    struct tm *ltm = localtime (&now);
    snprintf (buffer, sizeof (buffer), "0%02d%02d%05d%02d0103", ltm->tm_hour, ltm->tm_min,
              p->target, p->weight / 10);
    packit (data, buffer);
    transact (dev, data, 8, capture, 8);

    return 1;
}

/* get data */

int get_data(libusb_device_handle *dev, data_t *db)
{

    if (dev == NULL)
        return 0;

    int i;
    unsigned char data[8];
    unsigned char capture[6 * 18];

    for (i = 0; i < 18; i++) {

        memcpy (data, (const char[8]){ 6 * i, 0, 0, 0, 0, 0, 0, 2 }, 8);
        transact (dev, data, 8, capture + i * 6, 8);

    }

    time_t now = time (0);
    struct tm *ltm = localtime (&now);

    for (i = 0; i < 8; i++) {

        switch (i) {
        case 0:
            db[i].steps = unpackit (capture, 35, 1) * 10000 + unpackit (capture, 12, 4);
            db[i].distance = unpackit (capture, 25, 5);
            db[i].calories = unpackit (capture, 30, 5);
            db[i].walking_time = unpackit(capture, 44, 4) * 10 + unpackit (capture, 24, 1);
            break;
        case 1:
            db[i].steps = unpackit (capture, 67, 5);
            db[i].distance = unpackit (capture, 81, 3) * 100 + unpackit (capture, 60, 2);
            db[i].calories = unpackit (capture, 62, 5);
            db[i].walking_time = unpackit (capture, 76, 5);
            break;
        case 2:
            db[i].steps = unpackit (capture, 95, 1) * 10000 +  unpackit (capture, 72, 4);
            db[i].distance = unpackit (capture, 85, 5);
            db[i].calories = unpackit (capture, 90, 5);
            db[i].walking_time = unpackit(capture, 104, 4) * 10 + unpackit (capture, 84, 1);
            break;
        case 3:
            db[i].steps = unpackit (capture, 99, 5);
            db[i].distance = unpackit (capture, 113, 5);
            db[i].calories = unpackit (capture, 118, 2) * 1000 + unpackit (capture, 96, 3);
            db[i].walking_time = unpackit (capture, 108, 5);
            break;
        case 4:
            db[i].steps = unpackit (capture, 127, 5);
            db[i].distance = unpackit (capture, 141, 3) * 100 + unpackit (capture, 120, 2);
            db[i].calories = unpackit (capture, 122, 5);
            db[i].walking_time = unpackit (capture, 136, 5);
            break;
        case 5:
            db[i].steps = unpackit (capture, 155, 1) * 10000 + unpackit (capture, 132, 4);
            db[i].distance = unpackit (capture, 145, 5);
            db[i].calories = unpackit (capture, 150, 5);
            db[i].walking_time = unpackit (capture, 164, 4) * 10 + unpackit (capture, 144, 1);
            break;
        case 6:
            db[i].steps = unpackit (capture, 159, 5);
            db[i].distance = unpackit (capture, 173, 5);
            db[i].calories = unpackit (capture, 178, 2) * 1000 + unpackit (capture, 156, 3);
            db[i].walking_time = unpackit (capture, 168, 5);
            break;
        case 7:
            db[i].steps = unpackit (capture, 187, 5);
            db[i].distance = unpackit (capture, 201, 3) * 100 + unpackit (capture, 180, 2);
            db[i].calories = unpackit (capture, 182, 5);
            db[i].walking_time = unpackit (capture, 196, 5);
            break;
        }

        snprintf (db[i].date, sizeof (db[i].date), "%04d-%02d-%02d",
                  ltm->tm_year + 1900, ltm->tm_mon + 1, ltm->tm_mday);

        VERBOSE (DEBUG, printf ("date: %s step: %d distance: %d.%02d calories: %d.%01d walking time: %d\n",
                                db[i].date, db[i].steps, db[i].distance / 100, db[i].distance % 100,
                                db[i].calories / 10, db[i].calories % 10, db[i].walking_time));

        now -= 86400;
        ltm = localtime (&now);

    }

    return 1;
}

/* convert gait */

int convert_gait (system_t old_system, system_t new_system, int gait)
{
    switch (old_system - new_system) {
    case -1: gait = DIVROUND (gait, 3937); break;
    case 0: break;
    case 1: gait = DIVROUND (gait, 25400); break;
    }
    return gait;
}

/* convert weight */

int convert_weight (system_t old_system, system_t new_system, int weight)
{
    switch (old_system - new_system) {
    case -1: weight = DIVROUND (weight,  22046); break;
    case 0: break;
    case 1: weight = DIVROUND (weight, 4535); break;
    }
    return weight;
}

/* open db */

DB *dbp = NULL;

DB *open_db (void)
{
    static int init = 1;

    if (init) {
        if (db_create (&dbp, NULL, 0)) {
            VERBOSE (ERROR, printf ("Error initializing database structure\n"));
        } else if (dbp->open (dbp, NULL, dbfilename, NULL, DB_HASH, DB_CREATE, S_IRUSR | S_IWUSR)) {
            VERBOSE (ERROR, printf ("Error opening database '%s'\n", dbfilename));
            dbp = NULL;
        }
        init = 0;
    }
    return dbp;
}

/* close db */

/* help function */

void usage (int ret)
{
    FILE *fd = ret ? stderr : stdout;
    fprintf (fd, "usage: %s [-b db] [-d] [-f] [-g int] [-h] [-i] [-r] [-s int] [-t int] [-V] [-v int] [-w int] [-x] [-z int]\n",
             progname);
    fprintf (fd, " -b define Berkeley DB file name ($HOME/.%s.db)\n", progname);
    fprintf (fd, " -d fill database\n");
    fprintf (fd, " -f find device\n");
    fprintf (fd, " -g set gait (in cm or in)\n");
    fprintf (fd, " -h show help message\n");
    fprintf (fd, " -i get parameters\n");
    fprintf (fd, " -r retreive data from device\n");
    fprintf (fd, " -s set system (1 for metric, 2 for imperial)\n");
    fprintf (fd, " -t set target (in feet)\n");
    fprintf (fd, " -V show version\n");
    fprintf (fd, " -v verbose level (%d)\n", verbose);
    fprintf (fd, " -w set weight (in kg or lb)\n");
    fprintf (fd, " -x extract from db\n");
    fprintf (fd, " -z set db system (1 for metric, 2 for imperial)\n");

    exit (ret);
}

/* mode */

#define MODE_FIND_DEV    1
#define MODE_GET_PARAM   2
#define MODE_SET_PARAM   4
#define MODE_GET_DATA    8
#define MODE_FILL_DB    16
#define MODE_EXTRACT_DB 32
#define MODE_VERSION    64
#define MODE_CONVERT_DB 128

/* main function */

int main (int argc, char *argv[])
{

    libusb_device_handle *dev = NULL;
    parameter_t ped_params;
    parameter_t new_params = { 0, 0, 0, 0 };
    config_t db_config;
    config_t new_config = { 0 };
    data_t db[8];
    int i = 0;

    /* basename of program file */
    progname = argv[0];
    while (progname[i] != '\0') {
        if (progname[i++] == '/') {
            progname += i;
            i = 0;
        }
    }

    int mode = 0;

    int c;
    while ((c = getopt(argc, argv, "b:dfg:hirs:t:Vv:w:xz:")) != EOF) {
        switch (c) {
        case 'b':
            strncpy (dbfilename, optarg, sizeof (dbfilename));
            break;
	case 'd':
	    mode |= MODE_FIND_DEV;
	    mode |= MODE_GET_DATA;
	    mode |= MODE_FILL_DB;
	    break;
	case 'f':
	    mode |= MODE_FIND_DEV;
	    break;
        case 'g':
            mode |= MODE_FIND_DEV;
	    mode |= MODE_SET_PARAM;
            new_params.gait = atoi (optarg);
	    break;
        case 'h':
        default:
            usage (c != 'h');
	    break;
	case 'i':
	    mode |= MODE_FIND_DEV;
	    mode |= MODE_GET_PARAM;
            break;
	case 'r':
	    mode |= MODE_FIND_DEV;
	    mode |= MODE_GET_DATA;
	    break;
        case 's':
            mode |= MODE_FIND_DEV;
	    mode |= MODE_SET_PARAM;
            switch (atoi (optarg)) {
            case 1: new_params.system = system_metric_e; break;
            case 2: new_params.system = system_imperial_e; break;
            default: VERBOSE (WARNING, printf ("Unknown system (%s)\n", optarg));
            }
	    break;
        case 't':
            mode |= MODE_FIND_DEV;
	    mode |= MODE_SET_PARAM;
            new_params.target = atoi (optarg);
	    break;
        case 'V':
            mode |= MODE_VERSION;
            break;
        case 'v':
            verbose = atoi (optarg);
            break;
        case 'w':
            mode |= MODE_FIND_DEV;
	    mode |= MODE_SET_PARAM;
            new_params.weight = atoi (optarg);
	    break;
        case 'x':
            mode |= MODE_EXTRACT_DB;
            break;
        case 'z':
            mode |= MODE_CONVERT_DB;
            switch (atoi (optarg)){
            case 1: new_config.system = system_metric_e; break;
            case 2: new_config.system = system_imperial_e; break;
            default: VERBOSE (WARNING, printf ("Unknown system (%s)\n", optarg));
            }
            break;
        }
    }
    if (argc - optind != 0) {
        fprintf (stderr, "%s: invalid option -- %s\n", progname, argv[optind]);
        usage (1);
    }

    if (mode == 0) {
        VERBOSE (WARNING, printf ("Nothing to do\n"));
    }

    if (strlen (dbfilename) == 0) {
        snprintf (dbfilename, sizeof (dbfilename), "%s/.%s.db", getenv ("HOME"), progname);
    }

    /* Show version */
    if (mode & MODE_VERSION) {
        printf ("Version: %s%s" HGID "%s\n", VERSION, strlen (HGID) ? " (" : "", strlen (HGID) ? ")" : "");
    }

    /* Find device */
    if (mode & MODE_FIND_DEV) {
        VERBOSE (DEBUG, printf ("Looking for device\n"));
        dev = find_device ();
	if (dev) {
            VERBOSE (DEBUG, printf ("Device found\n"));
        } else {
            return 1;
        }
    }

    /* Get parameters */
    if (mode & MODE_GET_PARAM) {
        VERBOSE (DEBUG, printf ("Get parameters\n"));
        if (get_parameters (dev, &ped_params)) {
            printf ("Pedometer gait: %d%s\n", ped_params.gait, (ped_params.system == system_metric_e) ? "cm" : "in");
            printf ("Pedometer system: %s\n", (ped_params.system == system_metric_e) ? "metric" : "imperial");
            printf ("Pedometer target: %d feet\n", ped_params.target);
            printf ("Pedometer weight: %d%s\n", ped_params.weight, (ped_params.system == system_metric_e) ? "kg" : "lfb");
        }
    }

    /* Set parameters */
    if (mode & MODE_SET_PARAM) {
        VERBOSE (DEBUG, printf ("Set parameters\n"));
        if (get_parameters (dev, &ped_params)) {
            if (new_params.system == 0) new_params.system = ped_params.system;
            if (new_params.gait == 0)
                new_params.gait = convert_gait (ped_params.system, new_params.system, ped_params.gait);
            if (new_params.weight == 0)
                new_params.weight = convert_weight (ped_params.system, new_params.system, ped_params.weight);
            if (new_params.target == 0) new_params.target = ped_params.target;
            if (set_parameters (dev, &new_params)) {
                printf ("New gait: %d%s\n", new_params.gait, (new_params.system == system_metric_e) ? "cm" : "in");
                printf ("New system: %s\n", (new_params.system == system_metric_e) ? "metric" : "imperial");
                printf ("New target: %d feet\n", new_params.target);
                printf ("New weight: %d%s\n", new_params.weight, (new_params.system == system_metric_e) ? "kg" : "lfb");
            }
        }
    }

    /* Convert database */
    if (mode & MODE_CONVERT_DB) {
        VERBOSE (DEBUG, printf ("CONVERT database\n"));
        DB *dbp = open_db ();
        if (dbp) {
            DBT key;
            memset (&key, 0, sizeof (key));
            DBT data;
            memset (&data, 0, sizeof (data));
            char *keyname = "config";
            key.data = keyname;
            key.size = strlen (keyname) + 1;
            data.data = &db_config;
            data.ulen = sizeof (db_config);
            data.flags = DB_DBT_USERMEM;
            if (dbp->get (dbp, NULL, &key, &data, 0) == DB_NOTFOUND) {
                VERBOSE (ERROR, printf ("Config not found in database\n"));
                memset (&db_config, 0, sizeof (db_config));
            } else {
                VERBOSE (DEBUG, printf ("Database system: %u\n", db_config.system));
            }
            if (db_config.system == system_undef_e) {
                printf ("what is the database system (1 for metric, 2 for imperial)? ");
                if ((scanf("%u", &db_config.system)) && (db_config.system >= system_unknown_e)) {
                    VERBOSE (WARNING, printf ("Unknown system (%u)\n", db_config.system));
                    db_config.system = system_undef_e;
                }
                if (db_config.system != system_undef_e) {
                    data.size = sizeof (db_config);
                    if (dbp->put (dbp, NULL, &key, &data, DB_OVERWRITE_DUP)) {
                        VERBOSE (WARNING, printf ("Failed inserting config into database\n"));
                    } else {
                        VERBOSE (DEBUG, printf ("Insert initial db config\n"));
                    }
                }
            }
            VERBOSE (DEBUG, printf ("Config: %u -> %u\n", db_config.system, new_config.system));
            if (new_config.system) {
                if (new_config.system != db_config.system) {
                    data.data = &new_config;
                    data.size = sizeof (new_config);
                    if (dbp->put (dbp, NULL, &key, &data, DB_OVERWRITE_DUP)) {
                        VERBOSE (WARNING, printf ("Failed inserting config into database\n"));
                    } else {
                        VERBOSE (DEBUG, printf ("Update database configuration\n"));
                        if (db_config.system && new_config.system) {
                            /* convert db */
                            data_t d;
                            DBT key;
                            memset (&key, 0, sizeof (key));
                            key.flags = DB_DBT_MALLOC;
                            DBT data;
                            memset (&data, 0, sizeof (data));
                            data.data = &d;
                            data.ulen = sizeof (data_t);
                            data.flags = DB_DBT_USERMEM;
                            DBC *cursorp;
                            dbp->cursor(dbp, NULL, &cursorp, 0);
                            int nb = 0, ret;
                            while (!(ret = cursorp->get (cursorp, &key, &data, DB_NEXT))) {
                                if (strncmp (key.data, "config", 6) != 0) {
                                    d.distance = convert_gait (new_config.system, db_config.system, d.distance * 10);
                                    d.distance = DIVROUND (d.distance, 1000);
                                    if (dbp->put (dbp, NULL, &key, &data, DB_OVERWRITE_DUP))
                                        VERBOSE (WARNING, printf ("Failed updating data into database\n"));
                                    nb++;
                                }
                                free (key.data);
                            }
                            VERBOSE (DEBUG, printf ("Process %d records\n", nb));
                        }
                    }
                } else {
                    VERBOSE (WARNING, printf ("Nothing to do\n"));
                }
            } else {
                VERBOSE (WARNING, printf ("Incoherent unit systeme (db or request), nothing to do\n"));
            }
        }
    }

    /* Get data */
    if (mode & MODE_GET_DATA) {
        VERBOSE (DEBUG, printf ("Get data\n"));
        if (get_data (dev, db)) {
            if (mode & MODE_FILL_DB) {
                VERBOSE (DEBUG, printf ("Fill database\n"));
                DB *dbp = open_db ();
                if (dbp) {
                    for (i = 0; i < 8; i++) {
                        if (db[i].steps == 0)
                            continue;
                        DBT key;
                        memset (&key, 0, sizeof (key));
                        DBT data;
                        memset (&data, 0, sizeof (data));
                        key.data = db[i].date;
                        key.size = strlen (db[i].date) + 1;
                        data.data = db + i;
                        data.size = sizeof (data_t);
                        VERBOSE (DEBUG, printf ("Inserting key '%s' into db\n", db[i].date));
                        if (dbp->put (dbp, NULL, &key, &data, DB_OVERWRITE_DUP))
                            VERBOSE (WARNING, printf ("Failed inserting into database\n"));
                    }
                }
            } else {
                printf ("date steps distance calorie time\n");
                for (i = 0; i < 8; i++) {
                    if (db[i].steps == 0)
                        continue;
                    printf ("%s %d %d.%02d %d.%01d %d\n", db[i].date, db[i].steps, db[i].distance / 100,
                            db[i].distance % 100, db[i].calories / 10, db[i].calories % 10,
                            db[i].walking_time);
                }
            }
        }
    }

    /* Extract from database */
    if (mode & MODE_EXTRACT_DB) {
        VERBOSE (DEBUG, printf ("Extract from database\n"));
        DB *dbp = open_db ();
        if (dbp) {
            data_t d;
            DBT key;
            memset (&key, 0, sizeof (key));
            key.flags = DB_DBT_MALLOC;
            DBT data;
            memset (&data, 0, sizeof (data));
            data.data = &d;
            data.ulen = sizeof (data_t);
            data.flags = DB_DBT_USERMEM;
            DBC *cursorp;
            dbp->cursor(dbp, NULL, &cursorp, 0);
            data_t *table = NULL;
            int i, nb = 0, ret;
            while (!(ret = cursorp->get (cursorp, &key, &data, DB_NEXT))) {
                if (strncmp (key.data, "config", 6) != 0) {
                    table = (data_t *) realloc (table, (nb + 2) * sizeof (data_t));
                    for (i = 0; i < nb; i++)
                        if (strcmp (d.date, (table + i)->date) > 0)
                            break;
                    if (i < nb)
                        memmove (table + i + 1, table + i, (nb - i) * sizeof (data_t));
                    memcpy (table + i, &d, sizeof (data_t));
                    nb++;
                }
                free (key.data);
            }
            VERBOSE (DEBUG, printf ("nb: %d\n", nb));
            if (nb)
                printf ("date       steps dist. calor. time step/h\n");
            for (i = 0; i < nb; i++) {
                char tmp[32] = {0};
                if ((table + i)->walking_time)
                    snprintf (tmp, sizeof (tmp), "%.1f", (table + i)->steps / (float)(table + i)->walking_time);
                printf ("%s %5d %3d.%01d %4d.%01d %4d %6s\n", (table + i)->date, (table + i)->steps,
                        (table + i)->distance / 100, (table + i)->distance / 10 % 10,
                        (table + i)->calories / 10, (table + i)->calories % 10,
                        (table + i)->walking_time, (strlen (tmp)) ? tmp : "-");
            }

            free (table);
            if (ret != DB_NOTFOUND)
		VERBOSE (ERROR, printf ("Nothing found in database\n"));
            if (cursorp != NULL)
		cursorp->close (cursorp);
        }
    }

    if (dev)
        free (dev);

    if (dbp)
        dbp->close (dbp, 0);

    return 0;
}

/* test: {} -h */
/* test: {} -h | awk '/usage:/ { rc=1 } END { exit (1-rc) }' */
/* test: {} -_ 2> /dev/null | awk 'END { if (NR == 0) { exit(0) } else exit (1) }' */
/* test: {} -_ 2>&1 | awk '/usage:/ { rc=1 } END { exit (1-rc) }' */
/* test: {} -s 1 -w 83 -g 78 -t 10000 && {} -i | awk '/ 78cm/ || / metric/ || / 10000 feet/ || / 83kg/ { t++ } END { exit (t != 4) }' */
/* test: {} -s 1 -w 83 -g 78 -t 10000 && {} -g 123 && {} -i | awk '/ 123cm/ || / metric/ || / 10000 feet/ || / 83kg/ { t++ } END { exit (t != 4) }' */
/* test: {} -s 1 -w 83 -g 78 -t 10000 && {} -t 8976 && {} -i | awk '/ 78cm/ || / metric/ || / 8976 feet/ || / 83kg/ { t++ } END { exit (t != 4) }' */
/* test: {} -s 1 -w 83 -g 78 -t 10000 && {} -w 345 && {} -i | awk '/ 78cm/ || / metric/ || / 10000 feet/ || / 345kg/ { t++ } END { exit (t != 4) }' */
/* test: {} -s 1 -w 83 -g 78 -t 10000 && {} -s 2 && {} -i | awk '/ 31in/ || / imperial/ || / 10000 feet/ || / 183lfb/ { t++ } END { exit (t != 4) }' */
/* test: {} -s 1 -w 83 -g 78 -t 10000 && {} -s 2 && {} -s 1 && {} -i | awk '/ 79cm/ || / metric/ || / 10000 feet/ || / 83kg/ {t++} END { exit (t != 4) }' */
/* test: {} -r */
/* test: {} -d && test -f $HOME/.pedometer.db */
/* test: {} -x */
/* test: {} -V  */
